""" Here are allocated all the Proprietary name conventions"""
from abc import ABC, abstractmethod
from pathlib import Path
import datetime
import os

from .basics import conventions


def is_proprietary_din(file: Path, year_bounds=(1800, 2300)) -> bool:
    """
    --------------------------------------------------------------------------
    Return if the filename is in one of the accepted proprietary DIN
    conventions
    --------------------------------------------------------------------------
    """
    for prop_class in BaseProprietary.__subclasses__():
        if prop_class().is_din(file, year_bounds):  # type: ignore
            return True
    return False


def kdin_from_proprietary_din(file: Path, year_bounds=(1800, 2300)) -> Path:
    """
    --------------------------------------------------------------------------
    - Get the file path with the KDIN from the given proprietary DIN
    --------------------------------------------------------------------------
    """
    # pylint: disable=protected-access
    for prop_class in BaseProprietary.__subclasses__():
        init_class = prop_class()  # type: ignore
        if init_class.is_din(file, year_bounds):
            dtt_date = init_class.get_din(file, year_bounds)
            str_date = dtt_date.strftime(init_class._din_fmt)
            new_name = "".join(file.name.split(str_date))
            tail0, tail1 = "", ""
            if init_class._tail:
                tail0 = " " + init_class._tail
                tail1 = tail0 + " "
            if not new_name:
                new_name = conventions.date2kdin(dtt_date) + tail0
            elif new_name == file.suffix or new_name[0] == " ":
                new_name = conventions.date2kdin(dtt_date) + tail0 + new_name
            else:
                new_name = conventions.date2kdin(dtt_date) + tail1 + new_name
            new_path = file.parent.joinpath(new_name)
            return new_path
    return file


def rename_proprietary_din_file(file: Path, year_bounds=(1800, 2300)) -> Path:
    """
    --------------------------------------------------------------------------
    - Rename the proprietary file with KDIN and return the new file path.
    - If the renamed file already exists it does nothing and returns the
    original filename.
    --------------------------------------------------------------------------
    """
    new_path = kdin_from_proprietary_din(file, year_bounds)
    if new_path != file:
        if new_path.exists():
            return file
        os.rename(file, new_path)
    return new_path


class BaseProprietary(ABC):
    """
    --------------------------------------------------------------------------
    Base Class for every proprietary convention. For including new conventions
    just add a new subclass at the bottom of the file. See the already added
    ones (Like Google, Screenshot, Whatsapp...) as example.
    --------------------------------------------------------------------------
    - _tail: To include a trace of the proprietary convention in the name
    - _din_fmt: Format of the date-in-name to obtain
    - _conditions_ok(): To define here special conditions to accomplish
    - _clean_datename(): Return the datename clean (only static values)
    --------------------------------------------------------------------------
    - NOTE: Only the abstract methods must be overwritten
    --------------------------------------------------------------------------
    """
    @property
    @abstractmethod
    def _tail(self) -> str:
        """
        ----------------------------------------------------------------------
        This must include a text to add after the KDIN and before the rest of
        the name of the file (only to include if it is desired to know the
        source of the original name-format or convention of the file)
        e.g: IMG, Whatsapp, Nokia, Nikon, etc.
        ----------------------------------------------------------------------
        """
        return ""

    @property
    @abstractmethod
    def _din_fmt(self) -> str:
        """
        ----------------------------------------------------------------------
        This must return the DIN format according to its convention, it can
        include static parts of the name
        ----------------------------------------------------------------------
        """
        return ""

    @staticmethod
    @abstractmethod
    def _conditions_ok(file: Path) -> bool:
        """
        ----------------------------------------------------------------------
        Here will be included for each specific Proprietary convention the
        specific conditions to accomplish for considering it valid DIN
        ----------------------------------------------------------------------
        """
        return False

    @staticmethod
    @abstractmethod
    def _clean_datename(file: Path) -> str:
        """
        ----------------------------------------------------------------------
        Return the date-in-name without any DIN variable fields (statics can
        be maintained) information in the text
        ----------------------------------------------------------------------
        """
        return ""

    def get_din(self, file: Path, year_bounds=(1800, 2300)
                ) -> datetime.datetime:
        """
        ----------------------------------------------------------------------
        Return the file DIN according to the selected proprietary convention.
        - If the DIN is not found, datetime(1, 1, 1) is returned
        ----------------------------------------------------------------------
        """
        try:
            if not self._conditions_ok(file):
                raise ValueError
            clean_name = self._clean_datename(file)
            new_date = datetime.datetime.strptime(clean_name, self._din_fmt)
            if year_bounds[0] <= new_date.year <= year_bounds[1]:
                return new_date
            raise ValueError
        except ValueError:
            return datetime.datetime(1, 1, 1)

    def is_din(self, file: Path, year_bounds=(1800, 2300)) -> bool:
        """
        ----------------------------------------------------------------------
        Return if the file has in the name the selected proprietary convention
        ----------------------------------------------------------------------
        """
        return self.get_din(file, year_bounds).year != 1


class GooglePhotos(BaseProprietary):
    """GooglePhotos Proprietary Convention"""
    @property
    def _tail(self) -> str:
        return super()._tail

    @property
    def _din_fmt(self) -> str:
        return r"IMG_%Y%m%d_%H%M%S"

    @staticmethod
    def _conditions_ok(file: Path) -> bool:
        if len(file.name) >= 19:
            if file.name[:4] == "IMG_" and file.name[12] == "_":
                return True
        return False

    @staticmethod
    def _clean_datename(file: Path) -> str:
        return file.name[:19]


class Screenshot(BaseProprietary):
    """Screenshot Proprietary Convention"""
    @property
    def _tail(self) -> str:
        return super()._tail

    @property
    def _din_fmt(self) -> str:
        return r"%Y-%m-%d %H.%M.%S"

    @staticmethod
    def _conditions_ok(file: Path) -> bool:
        if len(file.name) >= 19:
            spacers0 = file.name[10] == " "
            spacers1 = file.name[4] == file.name[7] == "-"
            spacers2 = file.name[13] == file.name[16] == "."
            if spacers0 and spacers1 and spacers2:
                return True
        return False

    @staticmethod
    def _clean_datename(file: Path) -> str:
        return file.name[:19]


class Whatsapp(BaseProprietary):
    """Whatsapp Proprietary Convention"""
    @property
    def _tail(self) -> str:
        return super()._tail

    @property
    def _din_fmt(self) -> str:
        return r"WhatsApp Image %Y-%m-%d at %H.%M.%S"

    @staticmethod
    def _conditions_ok(file: Path) -> bool:
        if len(file.name) >= 37:
            tst_blk0 = file.name[:15] == "WhatsApp Image "
            spacers0 = file.name[19] == file.name[22] == "-"
            spacers1 = file.name[31] == file.name[34] == "."
            if tst_blk0 and spacers0 and spacers1:
                return True
        return False

    @staticmethod
    def _clean_datename(file: Path) -> str:
        return file.name[:37]


if __name__ == "__main__":
    # ========================================================================
    # For executing these examples remove the '.' in the '.infodtb' import
    # ========================================================================
    print("=" * 79)
    _TST = Path.cwd().joinpath("IMG_20210105_010203ckasa")
    print(GooglePhotos().is_din(_TST))
    print(GooglePhotos().get_din(_TST))
    # ========================================================================
    _TST = Path.cwd().joinpath("2021-01-05 01.02.03ckasa")
    print(Screenshot().is_din(_TST))
    print(Screenshot().get_din(_TST))
    # ========================================================================
    _TST = Path.cwd().joinpath("WhatsApp Image 2018-09-22 at 18.11.39sdgasg")
    print(Whatsapp().is_din(_TST))
    print(Whatsapp().get_din(_TST))
    # =======================================================================
    print("=" * 79)



"""logging tools"""
from pathlib import Path
import datetime
import logging


def progress(uni_value: float, tale_text2add="", print_result=False) -> str:
    """Simple progress to string in '%' converter"""
    txt = f"{uni_value * 100:>5.2f}% " + tale_text2add
    if print_result:
        print(txt)
    return txt


def get_fast_logger(name: str, base_path=Path.cwd(), tofile=True,
                    initialize=True, log_level="DEBUG") -> logging.Logger:
    """
    ----------------------------------------------------------------------
    Initialize a simple logger with its CMD and FILE <name.log> handlers
    (if enabled) and return its handler (logging.Logger())
    - name: Name of the logger
    - base_path: path of the logger if tofile=True (base_path/name.log)
    - tofile: If enabled the logs will be recorded in a file
    - initialize: If enabled initialization prints/messages will be shown
    ----------------------------------------------------------------------
    """
    log_file = base_path.joinpath(name + '.log')
    if tofile:
        with open(log_file, "a", encoding="utf-8") as fle:
            if initialize:
                curr_tme = datetime.datetime.now()
                fle.write("-" * 41 + "\n")
                fle.write("NEW EXECUTION: " + str(curr_tme) + "\n")
                fle.write("-" * 41 + "\n")

    logger = logging.getLogger(name)
    logger.setLevel(log_level)

    hlrs = [isinstance(x, logging.FileHandler) for x in logger.handlers]
    has_file_handlers = True in hlrs

    hlrs = [isinstance(x, logging.StreamHandler) for x in logger.handlers]
    has_cmd_handlers = True in hlrs

    log_formatter = logging.Formatter(
        "%(asctime)s | %(levelname)-5.5s | %(message)s")

    if not has_file_handlers and tofile:
        file_handler = logging.FileHandler(log_file)
        file_handler.setFormatter(log_formatter)
        file_handler.setLevel(log_level)
        logger.addHandler(file_handler)

    if not has_cmd_handlers:
        console_handler = logging.StreamHandler()
        console_handler.setFormatter(log_formatter)
        console_handler.setLevel(log_level)
        logger.addHandler(console_handler)

    if initialize:
        logger.info("Logger initialized in '%s.log'", logger.name)
    return logger



"""File with basic file management tools in python"""
from typing import List, Tuple, Optional
from logging import Logger
from pathlib import Path
import shutil
import os


def itername(file: Path, separator="-", idx=1) -> Path:
    """
    --------------------------------------------------------------------------
    iterate the filename adding '{separator}{count}' starting from {idx} if
    the file exists and return a new file[path] for avoid overwriting
    --------------------------------------------------------------------------
    """
    if not file.exists():
        return file

    new_filename = Path(file)
    sfix = new_filename.suffix
    clean_name = new_filename.stem

    while new_filename.exists():
        new_name = clean_name + separator + f"{idx}" + sfix
        new_filename = file.parent.joinpath(new_name)
        idx += 1
    return new_filename


def get_folders_tree(base_folder: Path, filter_scan: Tuple[str, ...] = ()
                     ) -> List[Path]:
    """
    ---------------------------------------------------------------------------
    Get all the folders-tree allocated in a given <base_folder>

    - base_folder: Base path to get the folders tree.
    - filter_scan: [optional] Include a list of pattern-names in to scan in the
                   base_folder such as [*1*, stuffs*, *A, etc...] and the rest
                   of the folders in base_folder will be skiped.

    - Structure Example
        BaseFolder
            | FolderLvl1_A
            | ...
            | FolderLvl1_Z
    ---------------------------------------------------------------------------
    """
    assert base_folder.is_absolute(), "'base_folder' must be an absolute path."

    if not filter_scan:
        full_tree = [Path(x[0]) for x in os.walk(base_folder)]
        full_tree.sort()
        full_tree.pop(full_tree.index(base_folder))
        return full_tree

    # Get the folders matching the patterns to be scanned
    folders2scan = []
    for kwd in filter_scan:
        folders2scan += list(base_folder.glob(kwd))

    # Get the initial tree including only the desired folders
    folders_tree = [Path(x[0]) for x in os.walk(base_folder)]
    folders_tree = [x for x in folders_tree if x in folders2scan]
    folders_tree.sort()

    # Get all the complete tree of the folders in origin
    full_tree = []
    for folder in folders_tree:
        full_tree += [Path(x[0]) for x in os.walk(folder)]
    full_tree.sort()
    return full_tree


def get_files_tree(folders_tree: List[Path], extensions: Tuple[str, ...] = (),
                   upper_lower=True) -> List[Path]:
    """
    --------------------------------------------------------------------------
    Get all the files (absolute paths) matching the desired extensions 'ext'
    in a list of directories (If <extensions=())> it scans for ALL the files)

    - extensions: Tuple['NEF', 'JPG', etc...] (extensions without the '.')
    - upper_lower: if enabled, it looks for the extension in both upper+lower
    ---------------------------------------------------------------------------
    """
    exts = ["." + x for x in extensions]
    if extensions:
        for extension in extensions:
            assert_txt = f"The extensions must not contain '.' <{extensions}>"
            assert extension[0] != ".", assert_txt
        if upper_lower:
            exts = [x.upper() for x in exts]
            exts = list(exts) + [x.lower() for x in exts]

    full_files = []
    for folder in folders_tree:
        files_found = [folder.joinpath(x) for x in folder.glob("*")]
        files_found = [x for x in files_found if x.is_file()]
        if exts:
            files_found = [x for x in files_found if x.suffix in exts]
        if files_found:
            full_files += files_found
    full_files.sort()
    return full_files


def get_folders_from_files(files_tree: List[Path]) -> List[Path]:
    """
    --------------------------------------------------------------------------
    - Get a list of ABSOLUTE folders for a given ABSOLUTE files tree
    --------------------------------------------------------------------------
    """
    paths2create = []
    for file in files_tree:
        assert file.is_absolute(), "'files_tree' must contain absolute paths."
        assert file.is_file(), "'files_tree' must contain only file paths."
        if file.parent not in paths2create:
            paths2create.append(file.parent)
    return paths2create


def replicate_folders_in_path(relative_dirs2create: List[Path],
                              destination_path: Path,
                              logger: Optional[Logger] = None,
                              log_header: str = "") -> List[Path]:
    """
    --------------------------------------------------------------------------
    Create all the RELATIVE folders in <relative_dirs2create> into the
    <destination_path>
    - 'logger' to include a process log
    - 'log_header' to add a header before the message log
    --------------------------------------------------------------------------
    """
    folders_created: List[Path] = []
    for rel_folder in relative_dirs2create:
        new_folder = destination_path.joinpath(rel_folder)
        if not new_folder.exists():
            os.makedirs(new_folder)
            folders_created.append(rel_folder)
            if logger is not None:
                hdr = log_header if log_header else "Folder created:"
                logger.info(hdr + " %s", rel_folder)
    return folders_created


def move_files2destination(files_relative_tree: List[Path],
                           src_parent_folder: Path,
                           dst_parent_folder: Path,
                           logger: Optional[Logger] = None,
                           log_header: str = "") -> List[Path]:
    """
    --------------------------------------------------------------------------
    Move all the relative files from the <source_parent_folder> to the
    <destination_parent_folder>. (this does not create intermediate folder
    if needed, for this use 'replicate_folders_in_path').
    - 'files_rel_tree' must be list of RELATIVE folders
    - 'src_parent_folder' must be the origin parent folder
    - 'dst_parent_folder' must be the destination parent folder
    - 'logger' to include a process log
    - 'log_header' to add a header before the message log
    - Returns a list with all files moved
    --------------------------------------------------------------------------
    """
    files_moved: List[Path] = []
    err1 = "All files to move must exist in Origin: "
    err2 = "All files to move must NOT exist in Destination: "
    for file in files_relative_tree:
        origin_file = src_parent_folder.joinpath(file)
        destiny_file = dst_parent_folder.joinpath(file)
        assert origin_file.is_file(), err1 + str(file)
        assert not destiny_file.exists(), err2 + str(file)
        shutil.move(origin_file, destiny_file)
        files_moved.append(file)
        if logger is not None:
            hdr = log_header if log_header else "File moved:"
            logger.info(hdr + " %s", file)
    return files_moved



"""
------------------------------------------------------------------------------
Here are allocated all the Kjmaro Conventions (Kiko/Francisco Mata Aroco)
------------------------------------------------------------------------------
Folder Date-In-Name (DIN) bounds
    - Separators
        - '-' separate the same date-fields
        - '_' separate two dates (For year intervals '-' can be also used)
    - Valid combination of folder names
        - YYYY abcdefg1
        - YYYY-YYYY abcdefg2
        - YYYY_YYYY abcdefg3
        - YYYY-MM abcdefg4
        - YYYY-MM-DD abcdefg5
        - YYYY-MM_MM abcdefg6
        - YYYY-MM-DD_DD abcdefg7
        - YYYY-MM_YYYY-MM abcdefg8
        - YYYY-MM-DD_MM-DD abcdefg9
        - YYYY-MM-DD_YYYY-MM-DD abcdefg10
------------------------------------------------------------------------------
Files Date-In-Name (DIN)
    - YYYYMMDD-HHMMSS*
    - eg: 20210102-201005
    - eg: 20210102-201005.jpg
    - eg: 20210102-201005 .jpg
    - eg: 20210102-201005 test.jpg
    - eg: 20130502-235959 test2.jpg
    - ...
------------------------------------------------------------------------------
Files Edition Date-In-Name (EDIN) <[files to be edited by scripts]>
    - *++YYYY-MM-DD+HH-MM-SS++*
    - eg: test++1999-01-16+12-21-02++one.jpg
    - eg: test++1999-01-16+12-21-02++.jpg
    - eg: ++1999-01-16+12-21-02++one.jpg
    - ...
------------------------------------------------------------------------------
Files with DIN and Date-To-Review (TRDIN) <[files to be manually reviewed]>
- This is because the program couldn't determinate a valid date of the file.
    - YYYYMMDD-HHMMSS(DTR)*
    - eg: 20210102-201005(DTR).jpg
    - eg: 20130502-235959(DTR) test2.jpg
    - ...
"""
from typing import Tuple
from pathlib import Path
import datetime


def get_folder_kdin_bounds(folder: Path, year_bounds=(1800, 2300)
                           ) -> Tuple[datetime.datetime, datetime.datetime]:
    """
    --------------------------------------------------------------------------
    Find the date bounds in the folder-name following the Kjmaro convention.
    It returns the initial date as t0 and the final date t1 as +1 Year or
    Month or Date (depending on the case).
    if not valid date is found it returns datetime(1, 1, 1), datetime(1, 1, 1)
    Bounds are given including the days indicated
        - for 2021       -> 2021-01-01 to 2022-01-01
        - for 2021-10-26 -> 2021-10-26 to 2021-10-27
        - for 2021-10_11 -> 2021-10-01 to 2021-12-01
    --------------------------------------------------------------------------
    - Kjmaro Convention
        > '-' separate the same date-fields
        > '_' separate two dates (For year intervals '-' can be also used)
    - Kjmaro Valid combination of folder names
        - YYYY abcdefg1
        - YYYY-YYYY abcdefg2
        - YYYY_YYYY abcdefg3
        - YYYY-MM abcdefg4
        - YYYY-MM-DD abcdefg5
        - YYYY-MM_MM abcdefg6
        - YYYY-MM-DD_DD abcdefg7
        - YYYY-MM_YYYY-MM abcdefg8
        - YYYY-MM-DD_MM-DD abcdefg9
        - YYYY-MM-DD_YYYY-MM-DD abcdefg10
    --------------------------------------------------------------------------
    """
    # pylint: disable=too-many-locals, too-many-statements, too-many-branches
    fields = folder.name.split()[0].strip()
    date0 = datetime.datetime(1, 1, 1)
    date1 = datetime.datetime(1, 1, 1)
    deltaday = datetime.timedelta(days=1)

    # CASE1: YYYY:
    if len(fields) == 4 and fields.isdigit():
        in_bounds = year_bounds[0] <= int(fields) <= year_bounds[1]
        if in_bounds:
            date0 = datetime.datetime(int(fields), 1, 1)
            date1 = datetime.datetime(int(fields) + 1, 1, 1)

    # CASE2&3: YYYY-YYYY or YYYY_YYYY
    mid_bar = len(fields.split("-")) == 2
    low_bar = len(fields.split("_")) == 2
    if len(fields) == 9 and (mid_bar or low_bar):
        year0_str, year1_str = fields.split("-" if mid_bar else "_")
        if year0_str.isdigit() and year1_str.isdigit():
            in_bounds0 = year_bounds[0] <= int(year0_str) <= year_bounds[1]
            in_bounds1 = year_bounds[0] <= int(year1_str) <= year_bounds[1]
            if in_bounds0 and in_bounds1:
                date0 = datetime.datetime(int(year0_str), 1, 1)
                date1 = datetime.datetime(int(year1_str) + 1, 1, 1)

    # CASE4: YYYY-MM
    if len(fields) == 7 and len(fields.split("-")) == 2:
        year0_str, mnth0_str = fields.split("-")
        if year0_str.isdigit() and mnth0_str.isdigit():
            in_bounds_yr = year_bounds[0] <= int(year0_str) <= year_bounds[1]
            in_bounds_mt = 1 <= int(mnth0_str) <= 12
            if in_bounds_yr and in_bounds_mt:
                date0 = datetime.datetime(int(year0_str), int(mnth0_str), 1)
                date1 = date0 + datetime.timedelta(days=31)
                date1 = datetime.datetime(date1.year, date1.month, 1)

    # CASE5: YYYY-MM-DD
    if len(fields) == 10 and len(fields.split("-")) == 3:
        year0_str, mnth0_str, day_str = fields.split("-")
        if year0_str.isdigit() and mnth0_str.isdigit() and day_str.isdigit():
            in_bounds_yr = year_bounds[0] <= int(year0_str) <= year_bounds[1]
            in_bounds_mt = 1 <= int(mnth0_str) <= 12
            if in_bounds_yr and in_bounds_mt:
                try:
                    date0 = datetime.datetime(int(year0_str), int(mnth0_str),
                                              int(day_str))
                    date1 = date0 + datetime.timedelta(days=1)
                except ValueError:
                    pass

    # CASE6: YYYY-MM_MM
    tst0 = len(fields.split("-")) == 2
    tst1 = len(fields.split("_")) == 2
    if len(fields) == 10 and tst0 and tst1:
        year0_str = fields.split("-")[0]
        mnth0_str = fields.split("-")[1].split("_")[0]
        mnth1_str = fields.split("-")[1].split("_")[1]
        if year0_str.isdigit() and mnth0_str.isdigit() and mnth1_str.isdigit():
            in_bounds_yr = year_bounds[0] <= int(year0_str) <= year_bounds[1]
            in_bounds_mt0 = 1 <= int(mnth0_str) <= 12
            in_bounds_mt1 = 1 <= int(mnth1_str) <= 12
            if in_bounds_yr and in_bounds_mt0 and in_bounds_mt1:
                date0 = datetime.datetime(int(year0_str), int(mnth0_str), 1)
                date1 = datetime.datetime(int(year0_str), int(mnth1_str), 1)
                date1 = date1 + datetime.timedelta(days=31)
                date1 = datetime.datetime(date1.year, date1.month, 1)

    # CASE7: YYYY-MM-DD_DD
    tst0 = len(fields.split("-")) == 3
    tst1 = len(fields.split("_")) == 2
    if len(fields) == 13 and tst0 and tst1:
        year0_str = fields.split("-")[0]
        mnth0_str = fields.split("-")[1]
        day0_str = fields.split("-")[2].split("_")[0]
        day1_str = fields.split("-")[2].split("_")[1]
        days_are_digis = day0_str.isdigit() and day1_str.isdigit()
        if year0_str.isdigit() and mnth0_str.isdigit() and days_are_digis:
            in_bounds_yr = year_bounds[0] <= int(year0_str) <= year_bounds[1]
            in_bounds_mt = 1 <= int(mnth0_str) <= 12
            if in_bounds_yr and in_bounds_mt:
                try:
                    date0 = datetime.datetime(int(year0_str), int(mnth0_str),
                                              int(day0_str))
                    date1 = datetime.datetime(int(year0_str), int(mnth0_str),
                                              int(day1_str))
                    date1 = date1 + deltaday
                except ValueError:
                    pass

    # CASE8: YYYY-MM_YYYY-MM
    tst0 = len(fields.split("-")) == 3
    tst1 = len(fields.split("_")) == 2
    if len(fields) == 15 and tst0 and tst1:
        year0_str = fields.split("_")[0].split("-")[0]
        mnth0_str = fields.split("_")[0].split("-")[1]
        year1_str = fields.split("_")[1].split("-")[0]
        mnth1_str = fields.split("_")[1].split("-")[1]
        years_aredigit = year0_str.isdigit() and year1_str.isdigit()
        mnths_aredigit = mnth0_str.isdigit() and mnth1_str.isdigit()
        if years_aredigit and mnths_aredigit:
            in_bounds_yr0 = year_bounds[0] <= int(year0_str) <= year_bounds[1]
            in_bounds_yr1 = year_bounds[0] <= int(year1_str) <= year_bounds[1]
            in_bounds_mt0 = 1 <= int(mnth0_str) <= 12
            in_bounds_mt1 = 1 <= int(mnth1_str) <= 12
            yr_in_bounds = in_bounds_yr0 and in_bounds_yr1
            mt_in_bounds = in_bounds_mt0 and in_bounds_mt1
            if yr_in_bounds and mt_in_bounds:
                date0 = datetime.datetime(int(year0_str), int(mnth0_str), 1)
                date1 = datetime.datetime(int(year1_str), int(mnth1_str), 1)
                date1 = date1 + datetime.timedelta(days=31)
                date1 = datetime.datetime(date1.year, date1.month, 1)

    # CASE9: YYYY-MM-DD_MM-DD
    tst0 = len(fields.split("-")) == 4
    tst1 = len(fields.split("_")) == 2
    if len(fields) == 16 and tst0 and tst1:
        year0_str = fields.split("_")[0].split("-")[0]
        mnth0_str = fields.split("_")[0].split("-")[1]
        day0_str = fields.split("_")[0].split("-")[2]
        mnth1_str = fields.split("_")[1].split("-")[0]
        day1_str = fields.split("_")[1].split("-")[1]
        mnths_aredigit = mnth0_str.isdigit() and mnth1_str.isdigit()
        days_aredigit = day0_str.isdigit() and day1_str.isdigit()
        if year0_str.isdigit() and mnths_aredigit and days_aredigit:
            in_bounds_yr0 = year_bounds[0] <= int(year0_str) <= year_bounds[1]
            in_bounds_mt0 = 1 <= int(mnth0_str) <= 12
            in_bounds_mt1 = 1 <= int(mnth1_str) <= 12
            if in_bounds_yr0 and in_bounds_mt0 and in_bounds_mt1:
                try:
                    date0 = datetime.datetime(int(year0_str), int(mnth0_str),
                                              int(day0_str))
                    date1 = datetime.datetime(int(year0_str), int(mnth1_str),
                                              int(day1_str))
                    date1 = date1 + deltaday
                except ValueError:
                    pass

    # CASE10: YYYY-MM-DD_YYYY-MM-DD
    tst0 = len(fields.split("-")) == 5
    tst1 = len(fields.split("_")) == 2
    if len(fields) == 21 and tst0 and tst1:
        year0_str = fields.split("_")[0].split("-")[0]
        mnth0_str = fields.split("_")[0].split("-")[1]
        day0_str = fields.split("_")[0].split("-")[2]
        year1_str = fields.split("_")[1].split("-")[0]
        mnth1_str = fields.split("_")[1].split("-")[1]
        day1_str = fields.split("_")[1].split("-")[2]
        years_aredigit = year0_str.isdigit() and year1_str.isdigit()
        mnths_aredigit = mnth0_str.isdigit() and mnth1_str.isdigit()
        days_aredigit = day0_str.isdigit() and day1_str.isdigit()
        if years_aredigit and mnths_aredigit and days_aredigit:
            in_bounds_yr0 = year_bounds[0] <= int(year0_str) <= year_bounds[1]
            in_bounds_yr1 = year_bounds[0] <= int(year1_str) <= year_bounds[1]
            in_bounds_mt0 = 1 <= int(mnth0_str) <= 12
            in_bounds_mt1 = 1 <= int(mnth1_str) <= 12
            years_in_bounds = in_bounds_yr0 and in_bounds_yr1
            mnths_in_bounds = in_bounds_mt0 and in_bounds_mt1
            if years_in_bounds and mnths_in_bounds:
                try:
                    date0 = datetime.datetime(int(year0_str), int(mnth0_str),
                                              int(day0_str))
                    date1 = datetime.datetime(int(year1_str), int(mnth1_str),
                                              int(day1_str))
                    date1 = date1 + deltaday
                except ValueError:
                    pass
    return date0, date1


def get_file_kdin(file: Path, year_bounds=(1800, 2300)) -> datetime.datetime:
    """
    --------------------------------------------------------------------------
    Get the file date-in-name following the Kjmaro convention. Returns the
    date found in the filename > For not found dates return datetime(1, 1, 1)
    --------------------------------------------------------------------------
    - Kjmaro Convention
        > YYYYMMDD-HHMMSS*
        > eg: 20210102-201005
        > eg: 20210102-201005.jpg
        > eg: 20210102-201005 .jpg
        > eg: 20210102-201005 test.jpg
        > eg: 20130502-235959 test2.jpg
        > eg: 20130502-235959(DTR) test2.jpg
        > ...
    --------------------------------------------------------------------------
    """
    date = datetime.datetime(1, 1, 1)
    filename = file.name
    if len(filename) < 15:
        return date

    if filename[8] == "-":
        try:
            year, mnth, day = filename[:4], filename[4:6], filename[6:8]
            hour, mnts, sec = filename[9:11], filename[11:13], filename[13:15]
            if year_bounds[0] <= int(year) <= year_bounds[1]:
                date = datetime.datetime(int(year), int(mnth), int(day),
                                         int(hour), int(mnts), int(sec))
        except ValueError:
            pass
        except TypeError:
            pass
    return date


def get_file_ekdin(file: Path, year_bounds=(1800, 2300)) -> datetime.datetime:
    """
    --------------------------------------------------------------------------
    Get the file edit-date-in-name following the Kjmaro convention. Returns
    the date found in the filename > Not found dates return datetime(1, 1, 1)
    --------------------------------------------------------------------------
    This convention is used to automate the date modification of a file. The
    date is written in the filename and this means that the user want to add
    this date to the file.
    --------------------------------------------------------------------------
    - Kjmaro Convention
        > *++YYYY-MM-DD+HH-MM-SS++*
        > eg: test++1999-01-16+12-21-02++one.jpg
        > eg: test++1999-01-16+12-21-02++.jpg
        > eg: ++1999-01-16+12-21-02++one.jpg
        > ...
    --------------------------------------------------------------------------
    """
    # pylint: disable=too-many-locals
    date = datetime.datetime(1, 1, 1)
    filename_lst = file.name.split("++")

    if len(filename_lst) == 3:
        datename = str(filename_lst[1])
        blkok = len(datename.split("-")) == 5
        pckok = len(datename.split("+")) == 2
        sepok = datename[10] == "+"
        lnedte = datename[4] == "-" and datename[7] == "-"
        lnehur = datename[13] == "-" and datename[16] == "-"

        if blkok and pckok and sepok and lnedte and lnehur:
            year, mth, day = datename[:4], datename[5:7], datename[8:10]
            hour, mnt, sec = datename[11:13], datename[14:16], datename[17:19]
            try:
                if year_bounds[0] <= int(year) <= year_bounds[1]:
                    date = datetime.datetime(int(year), int(mth), int(day),
                                             int(hour), int(mnt), int(sec))
            except ValueError:
                pass
            except TypeError:
                pass
    return date


def is_folder_kdin(folder: Path, year_bounds=(1800, 2300)) -> bool:
    """
    --------------------------------------------------------------------------
    returns if the folder matches the Kjmaro naming convention
    --------------------------------------------------------------------------
    """
    date0, _ = get_folder_kdin_bounds(folder, year_bounds)
    return date0 != datetime.datetime(1, 1, 1)


def is_file_kdin(file: Path, year_bounds=(1800, 2300)) -> bool:
    """
    --------------------------------------------------------------------------
    returns if the file matches the Kjmaro date-in-name convention
    --------------------------------------------------------------------------
    This convention defines the date of the file when its metadata-date can
    not be edited. It must be placed at the beginning of the name.
    - Note: For Kjmaro date-edit convention see 'is_edit_kdin()' which has a
            differentiated date format.
    --------------------------------------------------------------------------
    """
    return get_file_kdin(file, year_bounds).year != 1


def is_file_ekdin(file: Path, year_bounds=(1800, 2300)) -> bool:
    """
    --------------------------------------------------------------------------
    returns if the file matches the Kjmaro file edit-date-in-name convention
    --------------------------------------------------------------------------
    This convention is used to automate the date modification of the file. The
    date is written in the filename and this means that the user want to add
    this date to the file.
    --------------------------------------------------------------------------
    """
    return get_file_ekdin(file, year_bounds).year != 1


def is_file_trkdin(file: Path, year_bounds=(1800, 2300)) -> bool:
    """
    --------------------------------------------------------------------------
    returns if the file matches the Kjmaro file To-Review Date-In-Name
    convention
    --------------------------------------------------------------------------
    """
    is_kdin = is_file_kdin(file, year_bounds)
    kdin_txt = date2kdin(get_file_kdin(file, year_bounds))
    return is_kdin and (kdin_txt + "(DTR)") in file.name


def date2kdin(date: datetime.datetime) -> str:
    """Convert a datetime to KDIN format"""
    return date.strftime(r'%Y%m%d-%H%M%S')


def date2ekdin(date: datetime.datetime) -> str:
    """Convert a datetime to EKDIN (edit KDIN) format"""
    return date.strftime(r'++%Y-%m-%d+%H-%M-%S++')


def file_ekdin2clean(file: Path) -> Path:
    """
    --------------------------------------------------------------------------
    Remove the Kjmaro edit-date-in-name convention from the file
    - EditionConvention: [*++YYYY-MM-DD+HH-MM-SS++*]
    --------------------------------------------------------------------------
    """
    clean_filename = file.name.split("++")
    clean_filename.pop(1)
    new_path = file.parent.joinpath(("".join(clean_filename)).strip())
    if new_path.parent != file.parent:
        new_path = new_path.joinpath(" ")
    return new_path


def file_ekdin2kdin(file: Path, year_bounds=(1800, 2300)) -> Path:
    """
    --------------------------------------------------------------------------
    Convert the filename from Kjmaro Edit-DIN to DIN Convention
    > EditionDINConvention: [*++YYYY-MM-DD+HH-MM-SS++*]
    > DateInNameConvention: [YYYYMMDD-HHMMSS*]
    --------------------------------------------------------------------------
    """
    kdin_as_str = date2kdin(get_file_ekdin(file, year_bounds))
    clean_file = file_ekdin2clean(file)
    if clean_file.name == " ":
        new_name = kdin_as_str
    elif clean_file.name == file.suffix or clean_file.name[0] == " ":
        new_name = kdin_as_str + clean_file.name
    else:
        new_name = kdin_as_str + " " + clean_file.name
    return file.parent.joinpath(new_name)


def file_clean2trkdin(file: Path, date2add: datetime.datetime) -> Path:
    """
    --------------------------------------------------------------------------
    Convert the filename to Kjmaro TRDIN Convention (date to review)
    > DateInNameConventionToReview: [YYYYMMDD-HHMMSS(DTR)*]
    --------------------------------------------------------------------------
    """
    kdin_date = date2kdin(date2add)
    if not file.suffix and file.name[0] == ".":
        new_name = file.with_name(kdin_date + "(DTR)" + file.name)
    else:
        new_name = file.with_name(kdin_date + "(DTR) " + file.name)
    return new_name



"""Package information"""

# pylint: disable=line-too-long
__all__ = ["__title__", "__summary__", "__uri__", "__version__",
           "__author__", "__email__", "__license__", "__copyright__",
           "PYTHON_REQUIRES", "INSTALL_REQUIRES", "CLASSIFIERS"]


# Package title, version, short description and repository URL
__title__ = "kjmarotools"
__version__ = "0.1.0"
__summary__ = "Python tools for photo, media management and other purposes"
__uri__ = f"https://github.com/fjmaro/{__title__.capitalize()}"  # Github Projet capitalized

# Author, email, license and copyright
__email__ = ""
__author__ = "Francisco José Mata Aroco"
__license__ = "License :: OSI Approved :: GNU General Public License v3 or later (GPLv3+)"
__copyright__ = f"2022 {__author__}"

# Python and package requirements
PYTHON_REQUIRES = ">=3.9, <4"
INSTALL_REQUIRES: list = []

# PyPI classifiers with '__license__' included (https://pypi.org/classifiers/)
CLASSIFIERS = [__license__,
               "Topic :: Multimedia",
               "Intended Audience :: Developers",
               "Programming Language :: Python :: 3",
               "Development Status :: 3 - Alpha", ]



"""conversion tools"""
from typing import Tuple
import datetime


def datetime2str(date_time: datetime.datetime, sp0="-", sp1=" ", sp2=":"
                 ) -> str:
    """
    --------------------------------------------------------------------------
    Convert datetime to string in a flexible format configuring sp0/sp1/sp2.
    - YYYY<sp0>MM<sp0>DD<sp1>HH<sp2>MM<sp2>SS
    --------------------------------------------------------------------------
    """
    return date_time.strftime(f"%Y{sp0}%m{sp0}%d{sp1}%H{sp2}%M{sp2}%S")


def str2datetime(date_time: str, sp0="-", sp1=" ", sp2=":"
                 ) -> datetime.datetime:
    """
    --------------------------------------------------------------------------
    Convert string to datetime in a flexible format configuring sp0/sp1/sp2.
    - YYYY<sp0>MM<sp0>DD<sp1>HH<sp2>MM<sp2>SS
    --------------------------------------------------------------------------
    """
    return datetime.datetime.strptime(
        date_time, f"%Y{sp0}%m{sp0}%d{sp1}%H{sp2}%M{sp2}%S")


def deg2dms_zone(degrees=0.0, zones_pos_neg: Tuple[str, str] = ("+", "-")
                 ) -> Tuple[int, int, float, str]:
    """Convert degrees to Deg-min-sec-zone +/- or N/S or X/X"""
    mnt = 60.0 * (degrees - int(degrees))
    seg = 60.0 * (mnt - int(mnt))
    zne = zones_pos_neg[0] if degrees >= 0 else zones_pos_neg[1]
    return abs(int(degrees)), abs(int(mnt)), abs(seg), zne


def dms_zone2deg(dgs=0.0, mns=0.0, scs=0.0, zone_positive=True) -> float:
    """Convert Deg-min-sec-zone (zone=1 or -1)... to deg"""
    deg_out = dgs + mns / 60 + scs / 3600
    return deg_out if zone_positive else -deg_out



"""tools related with the operative system"""
from pathlib import Path
import datetime
import hashlib
import os


def get_file_create_date(filepath: Path) -> datetime.datetime:
    """get the file creation date"""
    return datetime.datetime.fromtimestamp(os.path.getctime(filepath))


def get_file_modify_date(filepath: Path) -> datetime.datetime:
    """get the file modification date"""
    return datetime.datetime.fromtimestamp(os.path.getmtime(filepath))


def get_file_access_date(filepath: Path) -> datetime.datetime:
    """get the file access date"""
    return datetime.datetime.fromtimestamp(os.path.getatime(filepath))


def set_file_modify_date(filepath: Path, date: datetime.datetime):
    """set the file modify date immediately"""
    st_atime = os.stat(filepath).st_atime
    os.utime(filepath, (st_atime, int(datetime.datetime.timestamp(date))))


def set_file_access_date(filepath: Path, date: datetime.datetime):
    """set the file access date immediately"""
    st_mtime = os.stat(filepath).st_mtime
    os.utime(filepath, (int(datetime.datetime.timestamp(date)), st_mtime))


def md5checksum(filepath: Path, buffer=2**20) -> str:
    """return the MD5 value of the file"""
    with open(filepath, "rb") as fle:
        hashmd5 = hashlib.md5()
        while buff := fle.read(buffer):
            hashmd5.update(buff)
        return hashmd5.hexdigest()



"""
------------------------------------------------------------------------------
Setup file for autoloading '_about.py' information in standard distributions.
Looks for the '_about.py' file under the second level of the project folder.
------------------------------------------------------------------------------
Copyright - Francisco José Mata Aroco (https://github.com/fjmaro)
See LICENSE.md file in this project for further details (setup.py v1.0.0)
------------------------------------------------------------------------------
"""
# pylint: disable=exec-used, line-too-long
import glob
import pathlib
from setuptools import setup, find_packages

here = pathlib.Path(__file__).parent.resolve()
aboutpath = [x for x in glob.glob(str(here) + "/*/*") if "_about.py" in x]

if len(aboutpath) == 0:
    FileNotFoundError(
        "SetupError: File '_about.py' not found in package level './*/*'")
elif len(aboutpath) > 1:
    FileExistsError(
        "SetupError: Only one '_about.py' file is allowed in the package")

readmefound = pathlib.Path(here / "README.md").is_file()
assert readmefound, "SetupError: 'README.md' file not found"

about: dict = {}
with open(aboutpath[0], "r", encoding='utf-8') as fhd:
    exec(fhd.read(), about)

setup(name=about['__title__'],
      version=about['__version__'],
      description=about['__summary__'],
      long_description=(here / "README.md").read_text(encoding="utf-8"),
      url=about['__uri__'],
      author=about['__author__'],
      author_email=about['__email__'],
      license=about['__license__'],
      packages=find_packages(),
      classifiers=about['CLASSIFIERS'],
      python_requires=about['PYTHON_REQUIRES'],
      install_requires=about['INSTALL_REQUIRES'])
